import logging
from typing import List, Optional, Union
from urllib.parse import urljoin

import requests
from requests.adapters import HTTPAdapter, Retry, RetryError
from requests.exceptions import HTTPError

from ..assets import EXPORTED_FIELDS, QlikAsset
from .constants import (
    ASSET_PATHS,
    RESPONSE_DICT_EXPECTED_MSG,
    RESPONSE_LIST_EXPECTED_MSG,
    REST_API_BASE_PATH,
    RETRY_BACKOFF_FACTOR,
    RETRY_COUNTS,
    RETRY_STATUSES,
)

logger = logging.getLogger(__name__)

Response = Union[dict, List[dict]]


def _session() -> requests.Session:
    session = requests.Session()
    retry = Retry(
        total=RETRY_COUNTS,
        backoff_factor=RETRY_BACKOFF_FACTOR,
        status_forcelist=RETRY_STATUSES,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("https://", adapter)
    return session


def _check_next_page_url(links: dict, current_page_url: str) -> Optional[str]:
    next_page = links.get("next")
    if not next_page:
        return None

    next_page_url = next_page["href"]
    if next_page_url == current_page_url:
        return None

    return next_page_url


def _asset_path(asset: QlikAsset) -> str:
    assert asset in ASSET_PATHS, f"{asset} is not available through REST-API"
    return ASSET_PATHS[asset]


class RestApiClient:
    """
    Client class to connect to Qlik REST API and retrieve Qlik assets

    API documentation: https://qlik.dev/apis/#rest
    """

    def __init__(
        self,
        server_url: str,
        api_key: str,
        except_http_error_statuses: Optional[List[int]] = None,
    ):
        self._server_url = server_url
        self._api_key = api_key
        self._session = _session()
        self._except_http_error_statuses = except_http_error_statuses or []
        self._authenticate()

    def _authenticate(self):
        auth_header = {"Authorization": "Bearer " + self._api_key}
        self._session.headers.update(auth_header)

    @property
    def server_url(self) -> str:
        """Returns attribute server url"""
        return self._server_url

    def _url(self, asset_path: str, app_id: Optional[str] = None) -> str:
        """
        Formats the full url with a specific asset path

        Optionally the asset path can be provided an `app_id` when the url must
        be scoped on an app. Beware that the asset_path must then be formattable
        with an "app_id" argument.
        """
        if app_id:
            asset_path = asset_path.format(app_id=app_id)

        path = REST_API_BASE_PATH + asset_path
        return urljoin(self._server_url, path)

    def _handle_http_error(self, error: HTTPError):
        status_code = error.response.status_code
        if status_code in self._except_http_error_statuses:
            logger.warning(error)
            return None
        raise error

    def _call(self, url: str) -> Optional[Response]:
        try:
            response = self._session.get(url)
            response.raise_for_status()
            return response.json()
        except RetryError as error:
            logger.warning(error)
            return None
        except HTTPError as error:
            return self._handle_http_error(error)

    def _pager(self, first_page_url: str) -> List[dict]:
        current_page_url = first_page_url

        data: List[dict] = []
        while current_page_url:
            response = self._call(current_page_url)
            if not response:
                return data
            assert isinstance(response, dict), RESPONSE_DICT_EXPECTED_MSG
            data.extend(response["data"])

            links = response["links"]
            next_page_url = _check_next_page_url(links, current_page_url)
            if not next_page_url:
                return data

            current_page_url = next_page_url
        return data

    def get(self, asset: QlikAsset) -> List[dict]:
        """
        Calls the route corresponding to the asset and returns the list of
        corresponding data
        """
        asset_path = _asset_path(asset)
        url = self._url(asset_path)
        data = self._pager(url)

        def _filter_fields(row: dict) -> dict:
            return {key: row.get(key) for key in EXPORTED_FIELDS[asset]}

        return [_filter_fields(row) for row in data]

    def get_with_scope(self, asset: QlikAsset, app_id: str) -> List[dict]:
        """
        Calls the route corresponding to the asset scoped on an app_id and
        returns the corresponding data
        """
        asset_path = ASSET_PATHS[asset]
        url = self._url(asset_path, app_id=app_id)
        response = self._call(url)
        if not response:
            return []
        assert isinstance(response, list), RESPONSE_LIST_EXPECTED_MSG
        return response

    def spaces(self) -> List[dict]:
        """
        Returns the list of Spaces

        doc: https://qlik.dev/apis/rest/spaces/#%23%2Fentries%2Fspaces-get
        """
        return self.get(QlikAsset.SPACES)

    def users(self) -> List[dict]:
        """
        Returns the list of Users

        doc: https://qlik.dev/apis/rest/users/#%23%2Fentries%2Fusers-get
        """
        return self.get(QlikAsset.USERS)

    def apps(self) -> List[dict]:
        """
        Returns the list of Apps

        doc: https://qlik.dev/apis/rest/items/#%23%2Fentries%2Fitems-get
        """
        return self.get(QlikAsset.APPS)

    def data_lineage(self, app_id: str) -> List[dict]:
        """
        Returns the data lineage for a given app

        doc: https://qlik.dev/apis/rest/apps/#%23%2Fentries%2Fapps%2F-appId%2Fdata%2Flineage-get
        """
        return self.get_with_scope(QlikAsset.LINEAGE, app_id=app_id)



import logging
from typing import Iterable, List, Optional, Tuple, Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    validate_baseurl,
    write_json,
    write_summary,
)
from .assets import QlikAsset
from .client import QlikClient
from .constants import API_KEY, BASE_URL

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: QlikClient,
) -> Iterable[Tuple[QlikAsset, Union[list, dict]]]:
    """Iterate over the extracted data from Qlik"""

    logger.info("Extracting SPACES from REST API")
    spaces = client.fetch(QlikAsset.SPACES)
    yield QlikAsset.SPACES, deep_serialize(spaces)

    logger.info("Extracting USERS from REST API")
    users = client.fetch(QlikAsset.USERS)
    yield QlikAsset.USERS, deep_serialize(users)

    logger.info("Extracting APPS from REST API")
    apps = client.fetch(QlikAsset.APPS)
    yield QlikAsset.APPS, deep_serialize(apps)

    logging.info("Extracting LINEAGE data from REST API")
    lineage = client.fetch(QlikAsset.LINEAGE, apps=apps)
    yield QlikAsset.LINEAGE, deep_serialize(lineage)

    logging.info("Extracting MEASURES data from JSON-RPC API")
    measures = client.fetch(QlikAsset.MEASURES, apps=apps)
    yield QlikAsset.MEASURES, deep_serialize(measures)


def extract_all(
    base_url: Optional[str] = None,
    api_key: Optional[str] = None,
    output_directory: Optional[str] = None,
    except_http_error_statuses: Optional[List[int]] = None,
) -> None:
    """
    Extract data from Qlik REST API
    Store the output files locally under the given output_directory
    """

    _output_directory = output_directory or from_env(OUTPUT_DIR)
    _base_url = validate_baseurl(base_url or from_env(BASE_URL))
    _api_key = api_key or from_env(API_KEY)

    client = QlikClient(
        server_url=_base_url,
        api_key=_api_key,
        except_http_error_statuses=except_http_error_statuses,
    )

    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts, base_url=_base_url)



import json
from contextlib import contextmanager
from urllib.parse import urlsplit

import websocket  # type: ignore

from .constants import WS_HOST_TEMPLATE


def _ws_host_from_server_url(server_url: str) -> str:
    """Formats the websocket host from the server url"""
    split = urlsplit(server_url)
    return WS_HOST_TEMPLATE.format(hostname=split.hostname)


class WebsocketConnection:
    """
    A WebsocketConnection is a websocket connection opened for a given Qlik
    <app_id> to call the Qlik Engine Api

    doc: https://qlik.dev/apis/json-rpc/qix
    """

    def __init__(self, server_url: str, api_key: str):
        self.ws = websocket.WebSocket()
        self.origin = server_url
        self.host_template = _ws_host_from_server_url(server_url)
        self.api_key = api_key

        self.session: dict = {}

    def connect(self, app_id: str) -> None:
        """Connects the websocket to a qlik app"""
        auth_header = {"Authorization": "Bearer " + self.api_key}
        app_host = self.host_template.format(app_id=app_id)
        self.ws.connect(app_host, header=auth_header, origin=self.origin)
        self.session = json.loads(self.ws.recv())

    def _send_call(self, call_msg: str):
        self.ws.send(call_msg)
        return self.ws.recv()

    def call(self, message: dict) -> dict:
        """
        Sends a JSON-RPC message through the websocket connection and returns
        the response as dict.
        """
        serialized = json.dumps(message)
        _response = self._send_call(serialized)
        return json.loads(_response)

    def close(self):
        """
        Closes websocket connection.
        """
        self.ws.close()


@contextmanager
def open_websocket(app_id: str, server_url: str, api_key: str):
    """
    Context manager over websocket to connect to a Qlik app and close the
    connection at the end.
    """
    ws = WebsocketConnection(server_url, api_key)
    ws.connect(app_id=app_id)
    try:
        yield ws
    finally:
        ws.close()



from ....utils import RetryStrategy
from ..errors import RateLimitResponseError

RETRY_EXCEPTIONS = (RateLimitResponseError, ConnectionError)
RETRY_COUNT = 2
RETRY_BASE_MS = 10_000
RETRY_JITTER_MS = 1_000
RETRY_STRATEGY = RetryStrategy.CONSTANT


CLIENT_NAME = "ModeAnalytics/API"



from typing import List, Optional, Set, Tuple

from google.cloud.bigquery import Client as GoogleCloudClient  # type: ignore
from google.oauth2.service_account import Credentials  # type: ignore

from ..abstract import SqlalchemyClient

BIGQUERY_URI = "bigquery://"

CREDENTIALS_INFO_KEY = "credentials_info"
PROJECT_ID_KEY = "project_id"


class BigQueryClient(SqlalchemyClient):
    """Connect to BigQuery and run SQL queries"""

    def __init__(
        self,
        credentials: dict,
        db_allowed: Optional[Set[str]] = None,
        db_blocked: Optional[Set[str]] = None,
    ):
        super().__init__(credentials)
        self._db_allowed = db_allowed
        self._db_blocked = db_blocked

    @staticmethod
    def name() -> str:
        return "BigQuery"

    def _keep_project(self, project: str) -> bool:
        if self._db_allowed and project not in self._db_allowed:
            return False
        if self._db_blocked and project in self._db_blocked:
            return False
        return True

    def _engine_options(self, credentials: dict) -> dict:
        return {
            CREDENTIALS_INFO_KEY: credentials,
        }

    def _build_uri(self, credentials: dict) -> str:
        return BIGQUERY_URI

    def _google_cloud_client(self) -> GoogleCloudClient:
        assert (
            CREDENTIALS_INFO_KEY in self._options
        ), "Missing BigQuery credentials in engine's options"
        credentials = self._options[CREDENTIALS_INFO_KEY]
        return GoogleCloudClient(
            project=credentials.get(PROJECT_ID_KEY),
            credentials=Credentials.from_service_account_info(credentials),
        )

    def _list_datasets(self) -> List:
        client = self._google_cloud_client()
        return [
            dataset
            for project in client.list_projects()
            for dataset in client.list_datasets(project.project_id)
            if self._keep_project(project.project_id)
        ]

    def get_projects(self) -> List[str]:
        """
        Returns distinct project_id available for the given GCP client
        """
        client = self._google_cloud_client()
        return [
            p.project_id
            for p in client.list_projects()
            if self._keep_project(p.project_id)
        ]

    def get_regions(self) -> Set[Tuple[str, str]]:
        """
        Returns distinct (project_id, region) available for the given GCP client
        """
        return set(
            (ds.project, ds._properties["location"])
            for ds in self._list_datasets()
        )

    def get_datasets(self) -> Set[Tuple[str, str]]:
        """
        Returns distinct (project_id, dataset_id) available for the given GCP client
        """
        return set((ds.project, ds.dataset_id) for ds in self._list_datasets())



"""
The Looker sdk only support configuration through file stored on the disk
This is an attempt to bypass this and configure the sdk using a config dict.
"""

from typing import Any, Dict, Optional

from looker_sdk.rtl import (
    auth_session,
    requests_transport,
    serialize,
    transport,
)
from looker_sdk.sdk import constants
from looker_sdk.sdk.api40 import methods as methods40
from looker_sdk.sdk.api40.models import (
    Dashboard,
    DashboardElement,
    DBConnection,
    Folder,
    Look,
    LookmlModel,
    LookmlModelExplore,
    LookmlModelExploreField,
    LookmlModelExploreFieldset,
    LookmlModelExploreJoins,
    LookmlModelNavExplore,
    Project,
    Query,
    User,
)
from typing_extensions import Protocol

from ..env import timeout_second


class Credentials:
    """ValueObject for the credentials"""

    def __init__(
        self,
        *,
        base_url: str,
        client_id: str,
        client_secret: str,
        timeout: Optional[int] = None,
        **_kwargs,
    ):
        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.timeout: int = timeout or timeout_second()

    def to_dict(self) -> Dict[str, Any]:
        attrs = ("base_url", "client_id", "client_secret", "timeout")
        return {attr: getattr(self, attr) for attr in attrs}


def init40(config: Credentials) -> methods40.Looker40SDK:
    """Default dependency configuration"""
    settings = ApiSettings(config=config, sdk_version=constants.sdk_version)
    settings.is_configured()
    transport = requests_transport.RequestsTransport.configure(settings)
    return methods40.Looker40SDK(
        auth_session.AuthSession(
            settings,
            transport,
            serialize.deserialize40,
            "4.0",
        ),
        serialize.deserialize40,
        serialize.serialize,
        transport,
        "4.0",
    )


class CastorApiSettings(transport.PTransportSettings, Protocol):
    """This is an intermediate class that is meant to be extended (for typing purpose)"""

    def read_config(self) -> Dict[str, Any]:
        raise NotImplementedError()


class ApiSettings(CastorApiSettings):
    """SDK settings with initialisation using a credential object instead of a path to a .ini file"""

    def __init__(self, config: Credentials, sdk_version: Optional[str] = ""):
        """Configure using a config dict"""
        self.config = config.to_dict()
        self.verify_ssl = True
        self.base_url = self.config.get("base_url", "")
        self.timeout = config.timeout
        self.headers = {"Content-Type": "application/json"}
        self.agent_tag = f"{transport.AGENT_PREFIX}"
        if sdk_version:
            self.agent_tag += f" {sdk_version}"

    def read_config(self) -> Dict[str, Any]:
        """Returns a serialization of the credentials"""
        return self.config



import logging
from typing import Callable, Iterator, List, Optional, Sequence, Tuple

from ....utils import Pager, PagerLogger
from ..env import page_size
from ..fields import format_fields
from .sdk import (
    Credentials,
    Dashboard,
    DBConnection,
    Folder,
    Look,
    LookmlModel,
    LookmlModelExplore,
    Project,
    User,
    init40,
)

logger = logging.getLogger(__name__)

COMMON_DASH_LOOK_FIELDS = (
    "id",
    "title",
    "description",
    "folder_id",
    "created_at",
    "deleted_at",
    "last_accessed_at",
    "last_viewed_at",
    "view_count",
    "user_id",
)
FOLDER_FIELDS = (
    "id",
    "name",
    "parent_id",
    "creator_id",
    "is_personal",
    "is_personal_descendant",
)

DASHBOARD_FIELDS = (
    *COMMON_DASH_LOOK_FIELDS,
    {
        "dashboard_elements": (
            "id",
            "type",
            {
                "query": ("model", "view", "fields"),
                "look": ("id", {"query": ("model", "view", "fields")}),
                "result_maker": ("id", {"query": ("model", "view", "fields")}),
            },
        )
    },
)

LOOK_FIELDS = (
    *COMMON_DASH_LOOK_FIELDS,
    "updated_at",
    "last_updater_id",
    "query_id",
    "is_run_on_load",
    "model",  # { id, label }
)

USER_FIELDS = ("id", "avatar_url", "display_name", "email", "is_disabled")
LOOKML_FIELDS = (
    "name",
    "label",
    "has_content",
    "project_name",
    {"explores": ("name", "description", "label", "hidden", "group_label")},
)

CONNECTION_FIELDS = (
    "name",
    "dialect",
    "pdts_enabled",
    "host",
    "database",
    "schema",
    "tmp_db_name",
    "jdbc_additional_params",
    "dialect_name",
    "created_at",
    "user_id",
    "example",
    "user_attribute_fields",
    "sql_writing_with_info_schema",
)

PROJECT_FIELDS = (
    "id",
    "name",
    "uses_git",
    "git_remote_url",
    "git_username",
    "git_production_branch_name",
    "use_git_cookie_auth",
    "git_password_user_attribute",
    "git_service_name",
    "git_application_server_http_port",
    "git_application_server_http_scheme",
    "pull_request_mode",
    "validation_required",
    "git_release_mgmt_enabled",
    "allow_warnings",
    "is_example",
)

EXPLORE_FIELD_FIELDS = (
    "description",
    "field_group_label",
    "field_group_variant",
    "hidden",
    "is_filter",
    "label",
    "label_short",
    "measure",
    "name",
    "parameter",
    "primary_key",
    "project_name",
    "scope",
    "sortable",
    "sql",
    "tags",  # string[]
    "type",
    "view",
    "view_label",
    "times_used",
)

EXPLORE_JOIN_FIELDS = (
    "name",
    "dependent_fields",  # string[]
    "fields",  # string[]
    "foreign_key",
    "from_",
    "outer_only",
    "relationship",
    "required_joins",  # string[]
    "sql_foreign_key",
    "sql_on",
    "sql_table_name",
    "type",
    "view_label",
)

EXPLORE_FIELDS = (
    "id",
    "description",
    "name",
    "label",
    "title",
    "scopes",  # string[]
    "project_name",
    "model_name",
    "view_name",
    "hidden",
    "sql_table_name",
    "group_label",
    "tags",  # string[]
    {
        "fields": {
            "dimensions": EXPLORE_FIELD_FIELDS,
            "measures": EXPLORE_FIELD_FIELDS,
        },
        "joins": EXPLORE_JOIN_FIELDS,
    },
)

# Model from looker
LOOKML_PROJECT_NAME_BLOCKLIST = ("looker-data", "system__activity")


OnApiCall = Callable[[], None]


class ApiPagerLogger(PagerLogger):
    def __init__(self, on_api_call: Optional[OnApiCall]):
        self._on_api_call = on_api_call

    def on_page(self, page: int, count: int):
        logger.info(f"Fetched page {page} / {count} results")
        self._on_api_call and self._on_api_call()

    def on_success(self, page: int, total: int):
        logger.info(f"All page fetched: {page} pages / {total} results")


class ApiClient:
    """Looker client"""

    def __init__(
        self, credentials: Credentials, on_api_call: OnApiCall = lambda: None
    ):
        self._sdk = init40(credentials)
        self._on_api_call = on_api_call
        self._logger = ApiPagerLogger(on_api_call)
        self.per_page = page_size()

    def folders(self) -> List[Folder]:
        """Lists folders of the given Looker account"""

        def _search(page: int, per_page: int) -> Sequence[Folder]:
            return self._sdk.search_folders(
                fields=format_fields(FOLDER_FIELDS),
                per_page=per_page,
                page=page,
            )

        return Pager(_search, logger=self._logger).all(per_page=self.per_page)

    def dashboards(self) -> List[Dashboard]:
        """Lists dashboards of the given Looker account"""

        def _search(page: int, per_page: int) -> Sequence[Dashboard]:
            return self._sdk.search_dashboards(
                fields=format_fields(DASHBOARD_FIELDS),
                per_page=per_page,
                page=page,
            )

        return Pager(_search, logger=self._logger).all(per_page=self.per_page)

    def _search_looks(self) -> List[Look]:
        """
        fetch looks via `search_looks`
        https://developers.looker.com/api/explorer/4.0/methods/Look/search_looks
        """

        def _search(page: int, per_page: int) -> Sequence[Look]:
            return self._sdk.search_looks(
                fields=format_fields(LOOK_FIELDS), per_page=per_page, page=page
            )

        logger.info("Use search_looks endpoint to retrieve Looks")
        return Pager(_search, logger=self._logger).all(per_page=self.per_page)

    def _all_looks(self) -> List[Look]:
        """
        fetch looks via `all_looks`
        https://castor.cloud.looker.com/extensions/marketplace_extension_api_explorer::api-explorer/4.0/methods/Look/all_looks
        """
        logger.info("Use all_looks endpoint to retrieve Looks")

        # No pagination : see https://community.looker.com/looker-api-77/api-paging-limits-14598
        return list(self._sdk.all_looks(fields=format_fields(LOOK_FIELDS)))

    def looks(self, all_looks_endpoint: Optional[bool] = False) -> List[Look]:
        """Lists looks of the given Looker account

        By default, uses the endpoint `search_looks` that allow pagination.
        If `all_looks_endpoint` parameter is set to True, it uses `all_looks` endpoint.
        """
        if all_looks_endpoint:
            return self._all_looks()

        return self._search_looks()

    def users(self) -> List[User]:
        """Lists users of the given Looker account"""

        def _search(page: int, per_page: int) -> Sequence[User]:
            # HACK:
            # We use verified_looker_employee=False (filter out Looker employees)
            # Else api returns an empty list when no parameters are specified
            return self._sdk.search_users(
                fields=format_fields(USER_FIELDS),
                per_page=per_page,
                page=page,
                verified_looker_employee=False,
            )

        return Pager(_search, logger=self._logger).all(per_page=self.per_page)

    def lookml_models(self) -> List[LookmlModel]:
        """Iterates LookML models of the given Looker account"""

        models = self._sdk.all_lookml_models(
            fields=format_fields(LOOKML_FIELDS)
        )

        logger.info("All LookML models fetched")
        self._on_api_call()

        return [
            model
            for model in models
            if model.project_name not in LOOKML_PROJECT_NAME_BLOCKLIST
        ]

    def explores(
        self, explore_names=Iterator[Tuple[str, str]]
    ) -> List[LookmlModelExplore]:
        """Iterates explores of the given Looker account for the provided model/explore names"""

        def _call(model_name: str, explore_name: str) -> LookmlModelExplore:
            explore = self._sdk.lookml_model_explore(model_name, explore_name)

            logger.info(f"Explore {model_name}/{explore_name} fetched")
            self._on_api_call()
            return explore

        return [
            _call(model_name, explore_name)
            for model_name, explore_name in explore_names
        ]

    def connections(self) -> List[DBConnection]:
        """Lists databases connections of the given Looker account"""

        connections = self._sdk.all_connections(
            fields=format_fields(CONNECTION_FIELDS)
        )

        logger.info("All looker connections fetched")
        self._on_api_call()

        return list(connections)

    def projects(self) -> List[Project]:
        """Lists projects of the given Looker account"""

        projects = self._sdk.all_projects(fields=format_fields(PROJECT_FIELDS))

        logger.info("All looker projects fetched")
        self._on_api_call()

        return list(projects)



from enum import Enum
from typing import Dict


class TableauEntity(Enum):
    """
    Entities which can be fetched from Tableau
    """

    WORKBOOK = "workbooks"
    USER = "users"
    PROJECT = "projects"
    USAGE = "views"
    WORKBOOK_TO_DATASOURCE = "workbooks_to_datasource"
    DATASOURCE = "datasources"
    CUSTOM_SQL_TABLE = "custom_sql_tables"
    CUSTOM_SQL_QUERY = "custom_sql_queries"


EXPORTED_FIELDS = {
    TableauEntity.WORKBOOK: (
        "id",
        "name",
        "description",
        "tags",
        "project_id",
        "created_at",
        "updated_at",
        "owner_id",
        "webpage_url",
    ),
    TableauEntity.PROJECT: (
        "id",
        "name",
        "description",
        "parent_id",
    ),
    TableauEntity.USER: (
        "id",
        "name",
        "email",
        "fullname",
        "site_role",
    ),
    TableauEntity.USAGE: (
        "workbook_id",
        "total_views",
    ),
}


class TableauGraphqlEntity(Enum):
    """
    Entities which can be fetched from Tableau
    """

    WORKBOOK_TO_DATASOURCE = "workbooks"
    DATASOURCE = "datasources"
    CUSTOM_SQL = "customSQLTables"

    """
    Fields which will be use for Tableau GraphQL API
    """

    FIELDS_WORKBOOK_TO_DATASOURCE = """
        luid
        id
        embeddedDatasources {
            id
            name
        }
    """

    FIELDS_DATASOURCE = """
        id
        name
        hasExtracts
        upstreamTables {
            id
            schema
            name
            fullName
            database {
                id
                name
                connectionType
            }
        }
    """

    FIELDS_CUSTOM_SQL_TABLE = """
        id
        name
        columns {
            referencedByFields {
                datasource {
                    ... on PublishedDatasource {
                        id
                    }

                    ... on EmbeddedDatasource {
                        id
                    }
                }
            }
        }
        """

    FIELDS_CUSTOM_SQL_QUERY = """
        id
        name
        query
        database {
            name
            connectionType
        }
        tables {
            name
        }
    """


QUERY_FIELDS: Dict[TableauEntity, Dict[str, TableauGraphqlEntity]] = {
    TableauEntity.WORKBOOK_TO_DATASOURCE: {
        "object_type": TableauGraphqlEntity.WORKBOOK_TO_DATASOURCE,
        "fields": TableauGraphqlEntity.FIELDS_WORKBOOK_TO_DATASOURCE,
    },
    TableauEntity.DATASOURCE: {
        "object_type": TableauGraphqlEntity.DATASOURCE,
        "fields": TableauGraphqlEntity.FIELDS_DATASOURCE,
    },
    TableauEntity.CUSTOM_SQL_TABLE: {
        "object_type": TableauGraphqlEntity.CUSTOM_SQL,
        "fields": TableauGraphqlEntity.FIELDS_CUSTOM_SQL_TABLE,
    },
    TableauEntity.CUSTOM_SQL_QUERY: {
        "object_type": TableauGraphqlEntity.CUSTOM_SQL,
        "fields": TableauGraphqlEntity.FIELDS_CUSTOM_SQL_QUERY,
    },
}



import logging
from typing import Dict, List

import tableauserverclient as TSC  # type: ignore

from ....utils import EntitiesType
from ..constants import SAFE_MODE_PAGE_SIZE
from ..entities import TableauEntity
from ..errors import TableauErrorCode
from ..types import PageReturn, ServerResponseError
from ..usage import compute_usage_views
from .client_utils import extract_entity

logger = logging.getLogger(__name__)


def _paginated_option(page_number: int) -> TSC.RequestOptions:
    """Set up the Paginated option for TSC.RequestOptions"""
    return TSC.RequestOptions(
        pagesize=SAFE_MODE_PAGE_SIZE, pagenumber=page_number
    )


def _next_usage_page(client, page_number: int) -> PageReturn:
    """
    Request views per page
    return Usages | ServerResponseError | TableauErrorCode
    """
    options = _paginated_option(page_number)
    try:
        all_usages_items, _ = client._server.views.get(options, usage=True)
        return all_usages_items, None

    except ServerResponseError as error:
        expected = TableauErrorCode.PAGE_NUMBER_NOT_FOUND
        if error.code == expected.value:
            return None, expected
        raise error

    except ServerResponseError as error:
        return None, error


def safe_mode_fetch_usage(client) -> EntitiesType:
    """
    Iterate throught each page
    Returns computed usages when page number is not found
    Log errors if ServerResponseError is return
    """
    list_usages: List[Dict] = []
    page_number: int = 0

    while True:
        page_number += 1
        usages, error = _next_usage_page(client, page_number)
        if error == TableauErrorCode.PAGE_NUMBER_NOT_FOUND:
            return compute_usage_views(list_usages)

        if error:
            logger.warning(error)
            client.errors.append(str(error))
            continue

        if not usages:
            continue

        new_usages = [
            extract_entity(usage, TableauEntity.USAGE) for usage in usages
        ]
        list_usages.extend(new_usages)



import logging

from .constants import MEASURES_SESSION_PARAMS, JsonRpcMethod
from .error import JsonRpcError
from .json_rpc import JsonRpcClient
from .websocket import open_websocket

logger = logging.getLogger(__name__)


def _handle(response: dict) -> int:
    """Returns the object Handle from the response payload"""
    return response["result"]["qReturn"]["qHandle"]


def _list_measures(client: JsonRpcClient, app_id: str) -> list:
    """
    Executes JSON-RPC messaging sequence to retrieve the list of measures
    for a given app

    references:
        - https://community.qlik.com/t5/Integration-Extension-APIs/How-to-ask-for-a-list-of-dimensions-and-measures-in-Qlik-Engine/td-p/1412156
        - https://help.qlik.com/en-US/sense-developer/May2021/Subsystems/EngineAPI/Content/Sense_EngineAPI/GenericObject/overview-generic-object.htm
    """
    response = client.send_message(
        method=JsonRpcMethod.OPEN_DOC,
        params=[app_id],
    )
    app_handle = _handle(response)

    response = client.send_message(
        method=JsonRpcMethod.CREATE_SESSION_OBJECT,
        handle=app_handle,
        params=list(MEASURES_SESSION_PARAMS),
    )
    session_handle = _handle(response)

    response = client.send_message(
        method=JsonRpcMethod.GET_LAYOUT,
        handle=session_handle,
    )
    return response["result"]["qLayout"]["qMeasureList"]["qItems"]


class EngineApiClient:
    """
    Engine API client is responsible to send the sequence of messages to
    get measures using JsonRpcClient and websocket connection.
    """

    def __init__(self, server_url: str, api_key: str):
        self.server_url = server_url
        self.api_key = api_key

    def measures(self, app_id: str, skip_access_denied: bool = True) -> list:
        """
        Opens a websocket and pass it to a JsonRpcClient to return the list
        of measures scoped on an app_id.

        By default, access_denied errors are skipped (logged as warning)
        """

        with open_websocket(
            app_id=app_id, server_url=self.server_url, api_key=self.api_key
        ) as websocket:
            json_rpc_client = JsonRpcClient(websocket=websocket)

            try:
                return _list_measures(json_rpc_client, app_id)
            except JsonRpcError as error:
                if skip_access_denied and error.is_access_denied:
                    logger.warning(error)
                    return []

                raise error



import base64

from Crypto.Cipher import AES
from Crypto.Hash import SHA512
from Crypto.Protocol.KDF import PBKDF2

KEY_HASH_ITERATION = 100000
KEY_HASH_SALT = b""
KEY_HASH_SIZE = 64
ENCRYPTED_KEY_START = 32
ENCRYPTED_MSG_SIGNATURE_END = -32


def _decrypt_key(key: str) -> bytes:
    hashed_key = PBKDF2(  # type: ignore
        key,
        KEY_HASH_SALT,
        KEY_HASH_SIZE,
        count=KEY_HASH_ITERATION,
        hmac_hash_module=SHA512,
    )
    return hashed_key[ENCRYPTED_KEY_START:]


def _unpad(s: bytes) -> bytes:
    return s[: -ord(s[len(s) - 1 :])]


def decrypt(raw: str, secret_key: str) -> str:
    """
    Decryption function that takes
     * a raw encrypted string
     * the associated secret key required to decode the message
     and returns the decrypted message.
    """
    enc = base64.b64decode(raw)
    key = _decrypt_key(secret_key)
    iv = enc[: AES.block_size]
    cipher = AES.new(key, AES.MODE_CBC, iv)
    enc_message = enc[AES.block_size : ENCRYPTED_MSG_SIGNATURE_END]
    return _unpad(cipher.decrypt(enc_message)).decode("utf-8")
